/*-------------------------------------------------------------------------
 * Created by Martin Gerhardt
 * 
 * This Arduino sketch is to drive a clock with DCF77 receiver and a stepper motor.
 * 
 * Version 2.0
 */

//-----Includes-------------------------------------------------------------
#include <TimeLib.h>
#include <DCF77.h>
#include <AccelStepper.h>
#include "utility/Utils.h"
using namespace Utils;
//-----Includes-------------------------------------------------------------

//-----ArduinoNano_Pin_Settings---------------------------------------------
#define DCF_PIN 3           // Connection pin to DCF 77 device
#define DCF_INTERRUPT 1    // Interrupt number associated with pin
#define dirPin 2
#define stepPin 5
#define enablePin 8
#define disable HIGH
#define enable LOW
#define hallHourPin 4   //definer the hall input pins
#define hallMinutePin 7
//-----ArduinoNano_Pin_Settings---------------------------------------------

//-----Drive_Settings-------------------------------------------------------
#define clockwhise HIGH
#define counterclockwhise LOW
#define stepsPerRevolution 200 //stept of motor per revolution
#define microsteps 16 //1 means no microsteppting -> full step
const float stepsPerHour = 3800.505; //number of full motor steps needed for one hour to pass (one revolution of the minute pointer) -> means without microstepping
#define SPR stepsPerRevolution*microsteps
#define resetHour 4   //time where you reset to
#define resetMinute 0
#define outOfSync 30 //minutes every 12h which dont cause the clock to reset
//-----Drive_Settings-------------------------------------------------------

//-----Global_Variables-----------------------------------------------------
time_t lastSyncTime =0;
long differenceSyncTime =0;
time_t prevDisplay = 0;          // when the digital clock was displayed
time_t time; //time_t stores the elapsed seconds since 1970
int shadowTimeHour;   //a shadow object of the pointers
int shadowTimeMinute;
int shadowTimeSecond;
int differenceTimeHour;
int differenceTimeMinute;
int differenceTimeSecond;
long stepsToSet;
volatile int stepIncrement = 0;
long timeout = 0;
int currentStepsToDrive = 0;
int syncTrigger = 0;
long timercompare = 59203;
time_t latestupdatedTime = 0;
time_t previousupdatedTime = 0;
int newTimeTrigger =0;
time_t correctedrealtime = 0;
float tuningfactor = 1.0;
//-----Global_Variables-----------------------------------------------------

//-----Create_Objects-------------------------------------------------------
DCF77 DCF = DCF77(DCF_PIN,DCF_INTERRUPT);  // Define a stepper and the pins it will use
AccelStepper stepper(AccelStepper::DRIVER, stepPin, dirPin);
//-----Create_Objects-------------------------------------------------------

//-----Functions------------------------------------------------------------
/* Functions for testing
void singleFullStep(bool stepdirection){
  digitalWrite(dirPin, stepdirection);
  for (int i = 0; i < microsteps; i++) {  //correction for microstepping
    digitalWrite(stepPin, LOW);
    delayMicroseconds(2000);
    digitalWrite(stepPin, HIGH);
    }
}

void singleStep(bool stepdirection){
  digitalWrite(dirPin, stepdirection);
    digitalWrite(stepPin, LOW);
    delayMicroseconds(100);
    digitalWrite(stepPin, HIGH);
}
*/

/*Funktion to reset clock.
 * The pointers turn until both reed contacts are freshly activated.
 * There is a timeout implemented.
 */
void resetClock(){
  digitalWrite(dirPin, LOW); //in case there were negative steps the direction pin has to be set again.
  long stepcount = 0;
  float currentspeed = 1;
  stepper.setSpeed((long)(currentspeed*microsteps));
  
  while(digitalRead(hallHourPin)&& stepcount<stepsPerHour*13*microsteps){// the hour pointer is already at the position lets say 12:03 then we go until its not activated anymore
   //run until hour pointer is 0
    if(stepper.runSpeed()){
      if(currentspeed < 1000){
        currentspeed = currentspeed +0.1;
        stepper.setSpeed((long)(currentspeed*microsteps));
        }
      stepcount++;
    }
  }

  while(!digitalRead(hallHourPin)&& stepcount<stepsPerHour*13*microsteps){
    //run until hour pointer is arrived
    if(stepper.runSpeed()){
      if(currentspeed < 1000){
        currentspeed = currentspeed +0.1;
        stepper.setSpeed((long)(currentspeed*microsteps));
        }
      stepcount++;
      }
  }

  while(!digitalRead(hallMinutePin)&&stepcount<stepsPerHour*13*microsteps){
    if(stepper.runSpeed()){
      if(currentspeed > 80){
        currentspeed = currentspeed - 0.3;
        stepper.setSpeed((long)(currentspeed*microsteps));
        }
      stepper.runSpeed();//run until minute pointer is arrived
      stepcount++;
    }
  }
  //stepper.setSpeed(1000*microsteps);

  if(stepcount>=(stepsPerHour*13*microsteps-1)){
    Serial.println("resetClock is overrunn!");
    while(1){}
    }     
  stepper.setCurrentPosition(0); //set to reset postion
  shadowTimeHour = resetHour;
  shadowTimeMinute = resetMinute;
  shadowTimeSecond = 0;
}

/*Function to calculate the current time difference between a timestamp and the shadow time. 
 * 
 */
long calcDifferenceTime(time_t timestamp){
    long differenceTime = 0;
    
    differenceTime = (hourFormat12(timestamp)*3600+minute(timestamp)*60+second(timestamp))-(shadowTimeHour*3600+shadowTimeMinute*60+shadowTimeSecond);
    if(differenceTime<=-21600){differenceTime = differenceTime +43200;} //add 12*3600, because we always go the shortest way
    if(differenceTime>21600){differenceTime = differenceTime -43200;} //sub 12*3600, because we always go the shortest way
    Serial.print("difference in sec.: ");Serial.println(differenceTime);
  
  return differenceTime;
}



/*Function to set the pointers to the current time from the reset position.
 * It calculates the steps needed to set the pointers. Therefore it calculates first the time difference 
 * from the reset time to the current time and then converts it to steps.
 */
void setPointerClock(){
  correctedrealtime = latestupdatedTime + (long)(tuningfactor *(now()-latestupdatedTime)); //calculate the corrected time. This is part of the tuning process.
  stepsToSet = calcDifferenceTime(correctedrealtime) * (stepsPerHour / 60.0 / 60.0);
  Serial.print("steps to set ");Serial.println(stepsToSet);
  stepIncrement=0;
  stepper.runToNewPosition(stepsToSet*microsteps);//set the pointers
  //here now we are lacking the time which passed while setting the pointers
  //this is covered by resetting the interrupt counter before
}

/*Function to set the clock.
 * It first resets the clock pointers then it waits until it gets a valid time from DCF77.
 * If a valid time is there the clock will be set.
 * It is called after startup and in case there is a big deviation while running.
 */
void setClocknow(){
  digitalWrite(enablePin, enable);
  resetClock();
  digitalWrite(enablePin, disable);
  while(timeStatus()== timeNotSet){//Wait until internal clock is set via DCF77
    delay(900);
    Serial.println("Waiting for DCF77 time ... ");
    digitalWrite(LED_BUILTIN, digitalRead(DCF_PIN));
    } 
  digitalWrite(enablePin, enable);
  setPointerClock();
  digitalWrite(enablePin, disable);
}
//-----Functions------------------------------------------------------------

//-----Interrupt Routine----------------------------------------------------
ISR(TIMER1_COMPA_vect){
  //OCR1A = timercompare;
  stepIncrement++;
}
//-----Interrupt Routine----------------------------------------------------

//-----Init-----------------------------------------------------------------
void setup()
{
  //----Pin_Config-------------------------
  pinMode(hallHourPin, INPUT); //Hall Pin input
  pinMode(hallMinutePin, INPUT);
  pinMode(stepPin, OUTPUT); //Motor driver Pins as output
  pinMode(dirPin, OUTPUT);
  pinMode(enablePin, OUTPUT);
  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, LOW);
  digitalWrite(stepPin, LOW); //initialize the motor pins
  digitalWrite(dirPin, LOW); 
  digitalWrite(enablePin, disable); //HIGH Disable
  //----Pin_Config-------------------------

  //----Motor_Config-----------------------
  stepper.setMaxSpeed(400*microsteps);
  stepper.setAcceleration(40*microsteps);
  stepper.setMinPulseWidth(20);
  stepper.setCurrentPosition(0);
  //----Motor_Config-----------------------
  
  //----Serial_Config----------------------
  Serial.begin(9600);
  //----Serial_Config----------------------
  
  //----Timer_Interrupt-------------------- 
  cli();//stop interrupts
  //set timer1 interrupt at 0,966976744s
  TCCR1A = 0;// set entire TCCR1A register to 0
  TCCR1B = 0;// same for TCCR1B
  TCNT1  = 0;//initialize counter value to 0
  // set compare match register according to excel calc
  OCR1A = timercompare;//  (must be <65536) and be sure there is enough headroom to the limits >5% for the correction mechanism
  // turn on CTC mode
  TCCR1B |= (1 << WGM12);
  // Set CS12 bits for 256 prescaler
  TCCR1B |= (1 << CS12);  
  // enable timer compare interrupt
  TIMSK1 |= (1 << OCIE1A);
  sei();//allow interrupts
  //----Timer_Interrupt-------------------- 
  
  //----DCF77_Config-----------------------
  DCF.Start();
  setSyncInterval(25);    //setup sync service for the internal clock
  setSyncProvider(getDCFTime);
  //----DCF77_Config-----------------------
  
  //when startup then reset the clock and set it with correct time
  setClocknow();
}
//-----Init-----------------------------------------------------------------

//-----Main_loop------------------------------------------------------------
void loop(){
  
  //for debug display the time
  if( now() != prevDisplay) //update the display only if the time has changed
  {
    prevDisplay = now();
    digitalClockDisplay();
  }

  //work off any steps
  currentStepsToDrive=stepIncrement;
  if(currentStepsToDrive!=0){
    digitalWrite(enablePin, enable);
    stepper.move(currentStepsToDrive*microsteps);
    //Serial.print("steps to drive ");Serial.println(currentStepsToDrive);
    while(stepper.run()){
      //timeout++;
      //if(timeout>=currentStepsToDrive*microsteps*2){Serial.println("overrun of normal move loop");}
    }
    digitalWrite(enablePin, disable);
    //timeout=0;
    stepIncrement=stepIncrement-currentStepsToDrive;
    }

  //display if there is reception on the DCF77 module
  digitalWrite(LED_BUILTIN, digitalRead(DCF_PIN));
  
  //check for reed contacts time and Sync
  if(digitalRead(hallMinutePin) && digitalRead(hallHourPin) && syncTrigger==0 && isAM()){  //if the hour pointer and minute pointer is in place first get new time and then sync the pointers. isAM shall make sure its done at night once a day

    //set Shadow time
    Serial.println("set shadow time");
    stepper.setCurrentPosition(0); //set to reset postion
    shadowTimeHour = resetHour;
    shadowTimeMinute = resetMinute;
    shadowTimeSecond = 0;

    //we wait here a while to have chance to catch a valid time signal
    while(timeout<9000 && newTimeTrigger==0){
      delay(100);
       //display if there is reception on the DCF77 module
      digitalWrite(LED_BUILTIN, digitalRead(DCF_PIN));
      timeout++;
    }
    if(newTimeTrigger==1){
      newTimeTrigger=0;
    }
    
    setPointerClock();    // set the clock again in case there is a new time signal. for instance summer winter time change
    timeout=0;
    syncTrigger=1;
  }
  if(digitalRead(hallMinutePin)==0 && digitalRead(hallHourPin)==0 && syncTrigger!=0){syncTrigger=0;}  //if the hour pointer and minute pointer is off place reset the trigger
}
//-----Main_loop------------------------------------------------------------

void digitalClockDisplay(){
  // digital clock display of the time
  Serial.println("");
  Serial.print(hour());
  printDigits(minute());
  printDigits(second());
  Serial.print(" ");
  Serial.print(day());
  Serial.print(" ");
  Serial.print(month());
  Serial.print(" ");
  Serial.print(year());
  Serial.println();
}

void printDigits(int digits){
  // utility function for digital clock display: prints preceding colon and leading 0
  Serial.print(":");
  if(digits < 10)
    Serial.print('0');
  Serial.print(digits);
}

unsigned long getDCFTime()
{
  time_t DCFtime = DCF.getTime();
  // Indicator that a time check is done
  if (DCFtime!=0) {
    previousupdatedTime = latestupdatedTime;
    latestupdatedTime = DCFtime;
    if(previousupdatedTime!=0){newTimeTrigger = 1;} // When we get a new valid time signal this trigger is set.
    Serial.println("X");

    differenceSyncTime = now()-latestupdatedTime;
    Serial.print("DifferenceSyncTime: ");Serial.println(differenceSyncTime); 
    if(differenceSyncTime>(-(outOfSync*60))&&differenceSyncTime<(+(outOfSync*60))){ //if the difference is in the tolerance then make the tuning of the time register
  
      /*here we calculate the tuning factor. the error drift of the internal time (arduino resonator) and the external atomic time (DCF77) is:
       *          (lastupdatetime-previousupdateTime)-Xalt*(now()-previousupdate)
       * error_%= ---------------------------------------------------------------
       *                      (lastupdatetime-previousupdateTime)
       *then the correction is:
       *
       *                 (lastupdatetime-previousupdateTime)-Xalt*(now()-previousupdate)
       * Xnew=Xalt+Xalt*---------------------------------------------------------------
       *                      (lastupdatetime-previousupdateTime)
       *
       */
      tuningfactor=((latestupdatedTime-previousupdatedTime)-tuningfactor*(now()-previousupdatedTime))/(latestupdatedTime-previousupdatedTime)*tuningfactor+tuningfactor;
      Serial.print("Tuningfactor: ");Serial.println(tuningfactor, 5); 
      // the relation between real passed time and the clock time is proportional:
      // Timepredivider             
      //------------------------ * CompareRegister * Stepsperhour * days*2*12hour = days*2*12hour * 60 * 60 
      //mainclockfrequency(16MHz)
      Serial.print("current timer compare ");Serial.println(OCR1A); 
      OCR1A = (long)(timercompare / tuningfactor);
      Serial.print("new timer compare tuned ");Serial.println(OCR1A);
    }
  }
  return DCFtime;
}
